#include "../../../include/GPU/gpu.hpp"
#include "SimpleHelper.hpp"
#include <limits.h>
#include <stdio.h>

namespace FileProcessor {
    using GPU::PositionU16;
    using GPU::SizeU16;

    struct SimpleTIMSize : private SimpleTIM {
        constexpr SimpleTIMSize() {
        }

        constexpr uint16_t getTextureWidth() const {
            return SimpleTIM::getTextureX();
        }

        constexpr uint16_t getTextureHeight() const {
            return SimpleTIM::getTextureY();
        }

        constexpr uint16_t getClutWidth() const {
            return SimpleTIM::getClutX();
        }

        constexpr uint16_t getClutHeight() const {
            return SimpleTIM::getClutY();
        }
    };

    struct SimpleTIMState {
        SimpleTIM     dst_info;
        SimpleTIMSize size_info;
        size_t        words_left; //32bit values

        constexpr SimpleTIMState(const SimpleTIM& dst_info) : dst_info(dst_info) {
        }
    };

    static void set_gpu_receive(const uint32_t* src, uint16_t x, uint16_t y, uint16_t w, uint16_t h) {
        GPU::DMA::Receive::prepare();
        GPU::DMA::Receive::set_dst(PositionU16(x, y), SizeU16(w, h)); 
        GPU::DMA::Receive::set_src(reinterpret_cast<const uintptr_t>(src));
    }

    static void set_gpu_receive_data(const uint32_t* src, SimpleTIMState& state, uint16_t width, uint16_t height) {
        state.words_left = (width*height)/2;
        set_gpu_receive(src, state.dst_info.getTextureX(), state.dst_info.getTextureY(), width, height);
    }

    static bool parse_data(State::Configuration& config, SimpleTIMState& state) {
        const auto words_to_use = (config.data_word_size > state.words_left) ? state.words_left : config.data_word_size;
        bool       is_last      = (words_to_use == state.words_left);
        auto       block_count  = (words_to_use >> 4);        

        while(block_count > 0) {
            const auto block_send = (block_count > UI16_MAX) ? UI16_MAX : block_count;

            // Send data!
            GPU::DMA::wait();
            GPU::DMA::Receive::start(block_send);
            block_count -= block_send;
        }

        if(is_last) {
            // Send words
            const auto last_words = (words_to_use & 0b1111);
            if(last_words > 0) {
                GPU::DMA::wait();
                GPU::DMA::Receive::start(1, last_words);
            }

            GPU::DMA::wait();
            GPU::DMA::end();

            state.words_left = 0;
            config.processed(words_to_use);

            return false;
        }

        else {
            const auto words_used = (words_to_use & ~0b1111);
            
            state.words_left -= words_used;
            config.processed(words_used);
            return true;
        }
    }

    static bool switch_state_parse_data(State::Configuration& config, SimpleTIMState& state) {
        set_gpu_receive_data(config.data_adr, state, state.size_info.getTextureWidth(), state.size_info.getTextureHeight());
        return Helper::exchange_and_execute_process_function(parse_data, config, state);
    }

    static bool parse_clut(State::Configuration& config, SimpleTIMState& state) {
        if(parse_data(config, state)) {
            return true;
        }

        else {
            return switch_state_parse_data(config, state);
        }
    }

    static bool parse_header(State::Configuration& config, SimpleTIMState& state) {
        if(config.data_word_size >= (sizeof(SimpleTIMSize)/sizeof(uint32_t))) {
            Helper::simple_read(state.size_info, config);
            
            //Check if we have a clut to care about
            if(state.size_info.getClutWidth() > 0) {
                //CLUTs are 16bit full color anyway
                set_gpu_receive_data(config.data_adr, state, state.size_info.getClutWidth(), state.size_info.getClutHeight());
                return Helper::exchange_and_execute_process_function(parse_clut, config, state);
            }

            //We have direct data
            else {
                return switch_state_parse_data(config, state);
            }
        }

        return true;
    }

    State create(const uint32_t* data_adr, const SimpleTIM& file) {        
        return State::from(SimpleTIMState(file), data_adr, parse_header);
    }
}

#undef private