use super::super::types::OverlaySlot;
use tool_helper::{Error, format_if_error, Output};
use std::io::Write;

const OVERLAY_DEFAULT_START:&'static str = "__boot_loader_end"; // < will probably be "__boot_loader_start" later

pub fn write(output: &mut Output, overlay_desc: &Vec<OverlaySlot>) -> Result<(), Error> {
    fn create_heap_base(overlay_desc: &Vec<OverlaySlot>) -> String {
        let mut heap_start = format!("MAX({}, __{}_end)", OVERLAY_DEFAULT_START, overlay_desc[0].name);

        for slot in overlay_desc.iter().skip(1) {
            heap_start = format!("MAX(__{}_end, {})", slot.name, heap_start);
        }
        heap_start
    }

    fn get_slot_start_adr(slot: &OverlaySlot) -> &str {
        if let Some(slot_adr) = &slot.start_adr {
            slot_adr.as_ref()
        }

        else {
            OVERLAY_DEFAULT_START
        }
    }

    if overlay_desc.is_empty() {
        return write_heap_base(output, OVERLAY_DEFAULT_START);
    }

    write_heap_base(output, &create_heap_base(overlay_desc))?;
    writeln!(output, "")?;
    writeln!(output, "SECTIONS {{")?;
        for slot in overlay_desc {
            writeln!(output, "\t/*{}*/", slot.name)?;
            writeln!(output, "\tOVERLAY {} : NOCROSSREFS SUBALIGN(4) {{", get_slot_start_adr(slot))?;
                for section in &slot.sections {
                    writeln!(output, "\t\t.{} {{", section.name)?;
                        writeln!(output, "\t\t\t__{}_start = .;", section.name)?;
                        section.file_pattern.iter().try_for_each(|patr| writeln!(output, "\t\t\tKEEP({}(.header))", patr))?;
                        writeln!(output, "\t\t\t__{}_lbas = .;", section.name)?;
                        section.file_pattern.iter().try_for_each(|patr| writeln!(output, "\t\t\tKEEP({}(.header.lbas))", patr))?;
                        writeln!(output, "\t\t\t__{}_ctor = .;", section.name)?;

                        for section_type in [".text.startup._GLOBAL__*", ".ctors", ".text.*", ".rodata*", ".sdata*", ".data*", ".sbss*", ".bss*"] {
                            section.file_pattern.iter().try_for_each(|patr| writeln!(output, "\t\t\t{}({})", patr, section_type))?;
                        }
                    writeln!(output, "\t\t}}")?;
                }
            writeln!(output, "\t}}")?;
            writeln!(output, "\t__{}_end = .;", slot.name)?;
            writeln!(output, "")?;
        }
    writeln!(output, "\t/DISCARD/ : {{ *(.comment*) *(.gnu.*) *(.mdebug*) *(.debug_*) *(.pdr)}}")?;
    writeln!(output, "}}")?;
    return Ok(());
}

fn write_heap_base(output: &mut Output, heap_base: &str) -> Result<(), Error> {
    format_if_error!(writeln!(output, "__heap_base = {};", heap_base), "Writing default LD Script failed with: {error_text}")?;
    Ok(())
}