use super::{OverlaySection, OverlaySlot};
use tool_helper::{Error, format_if_error, Input};

const FOLLOW_KEYWORD:&'static str = "following";

pub fn read_config(input: Input) -> Result<Vec<OverlaySlot>, Error> {
    let json_reader:serde_json::Value = format_if_error!(serde_json::from_reader(input), "Parsing JSON failed with: {error_text}")?;
    let mut overlay_slots             = Vec::<OverlaySlot>::new();

    if let Some(objects) = json_reader.as_object() {
        for object in objects {
            let mut slot = read_as_slot(object)?;
            if slot.start_adr.is_none() {
                if let Some(prev_slot) = overlay_slots.last() {
                    slot.set_start_adr_as_end_of(prev_slot.name.as_ref());
                }
            }

            overlay_slots.push(slot);
        }

        Ok(overlay_slots)
    }

    else {
        Err(Error::from_str("JSON must start with an object"))
    }
}

fn read_as_slot(json_value: (&String, &serde_json::Value)) -> Result<OverlaySlot, Error> {
    let (slot_name, json_value) = json_value;
    let mut slot                = OverlaySlot::new(slot_name.clone(), None);

    if let Some(json_value) = json_value.as_object() {
        for (name, json_value) in json_value {
            if name == FOLLOW_KEYWORD {
                if let Some(start_adr) = json_value.as_str() {
                    slot.set_start_adr_as_end_of(start_adr);
                }

                else {
                    return Err(Error::from_text(format!("\"{}\" option of {} must be a string", FOLLOW_KEYWORD, slot_name)));
                }
            }

            else {
                slot.add_section(read_as_section(name.clone(), json_value)?);
            }
        }
    }

    Ok(slot)
}

fn read_as_section(section_name: String, json_value: &serde_json::Value) -> Result<OverlaySection, Error> {
    fn create_pattern_error<T>(section_name: &String) -> Result<T, Error> {
        Err(Error::from_text(format!("Pattern element of section declaration \"{}\" must contain only strings", section_name)))
    }

    fn read_pattern(section_name: &String, json_value: &serde_json::Value) -> Result<String, Error> {
        if let Some(pattern) = json_value.as_str() {
            Ok(pattern.to_owned())
        }

        else {
            create_pattern_error(section_name)
        }
    }

    let mut section = OverlaySection::new(section_name);

    if let Some(pattern_json) = json_value.get("pattern") {
        if let Some(pattern) = pattern_json.as_array() {
            for cur_pattern in pattern {
                section.add_file_pattern(read_pattern(&section.name, cur_pattern)?);
            }
            Ok(section)
        }
        
        else if let Some(pattern) = pattern_json.as_str() {
            section.add_file_pattern(pattern.to_owned());
            Ok(section)
        }
        
        else {
            create_pattern_error(&section.name)
        }
    }
    
    else {
        Err(Error::from_text(format!("Section declaration \"{}\" does not contain any pattern!", section.name)))
    }
}