use cdtypes::{Error, types::{cue::*, helper::{sector_count_mode2_form1, sector_count_mode2_form2, round_bytes_mode2_form1}, dir_record::DirectoryRecord, sector::AudioSample}};
pub use super::types::{ProcessedFile, RequestedFile};

pub trait LBASize {
    fn set_lba(&mut self, lba: usize);
    fn set_lba_ret(&mut self, lba: usize) -> Result<usize, Error> {
        self.set_lba(lba);
        
        Ok(lba + self.get_length_lba()?)
    }
    fn get_lba(&self) -> usize;

    fn is_xa_audio(&self) -> Option<bool>;

    fn get_length(&self)     -> usize;
    fn get_length_lba(&self) -> Result<usize, Error> {
        self.default_get_length_lba()
    }

    //Rust doesn't support super calls yet
    fn default_get_length_lba(&self) -> Result<usize, Error> {
        let mut length = self.get_length();

        if length == 0 {
            length = 1;
        }

        if let Some(is_xa_audio) = self.is_xa_audio() {
            if is_xa_audio {
                Ok(sector_count_mode2_form2(length))
            }

            else {
                Ok(sector_count_mode2_form1(length))
            }
        }

        else {
            Err(Error::GenericError("Failed determining Sector type for LBA length calculation".to_owned()))
        }
    }
}

pub struct CDDesc {
    pub licence_path: String,
    pub output_file:  String,
    pub pvd:          PVD,
    pub tracks:       Vec<Track>,
}

impl CDDesc {
    pub fn new() -> CDDesc {
        CDDesc{licence_path: String::new(), output_file: String::new(), pvd: PVD::new(), tracks: Vec::new()}
    }
}

#[derive(Debug)]
pub struct PVD {
    pub volume_identifier: String,
    pub publisher:         String,
    pub data_preparer:     String,
}

impl PVD {
    pub fn new() -> PVD {
        PVD{volume_identifier: String::new(), publisher: String::new(), data_preparer: String::new()}
    }
    
    pub fn to_uppercase(&mut self) {
        self.volume_identifier = self.volume_identifier.to_uppercase();
        self.publisher         = self.publisher.to_uppercase();
        self.data_preparer     = self.data_preparer.to_uppercase();
    }
}

pub enum Track {
    Data(DirectoryDataType),
    Audio(Vec<AudioSample>),    
}

#[derive(Clone)]
pub enum DataTrackType {
    Directory(DirectoryDataType),
    File(FileDataType)
}

impl DataTrackType {
    pub fn get_name(&self) -> &String {
        match self {
            DataTrackType::Directory(dir) => &dir.name,
            DataTrackType::File(file)     => &file.name,
        }
    }

    pub fn is_hidden(&self) -> bool {
        match self {
            DataTrackType::Directory(dir) => dir.is_hidden,
            DataTrackType::File(file)     => file.is_hidden,
        }
    }
}

impl LBASize for DataTrackType {
    fn set_lba(&mut self, lba: usize) {
        match self {
            DataTrackType::Directory(dir) => dir.set_lba(lba),
            DataTrackType::File(file)     => file.set_lba(lba),
        }
    }

    fn get_lba(&self) -> usize {
        match self {
            DataTrackType::Directory(dir) => dir.get_lba(),
            DataTrackType::File(file)     => file.get_lba(),
        }
    }

    fn get_length(&self) -> usize {
        match self {
            DataTrackType::Directory(dir) => dir.get_length(),
            DataTrackType::File(file)     => file.get_length(),
        }
    }
    
    fn is_xa_audio(&self) -> Option<bool> {
        match self {
            DataTrackType::Directory(dir) => dir.is_xa_audio(),
            DataTrackType::File(file)     => file.is_xa_audio(),
        }
    }
}

#[derive(Clone)]
pub struct DirectoryDataType {
    pub name:      String,
        member:    Vec<DataTrackType>,
        lba:       usize,
        length:    usize,
    pub is_hidden: bool,
}

impl DirectoryDataType {
    const CUR_DIR_NAME:&'static str    = "\x00";
    const PARENT_DIR_NAME:&'static str = "\x01";

    pub fn new(name: &str) -> DirectoryDataType {
        DirectoryDataType{name: name.to_string(), member: Vec::new(), lba: 0usize, length: 0usize, is_hidden: false}
    }

    pub fn new_pvd_root_dir(lba: usize) -> DirectoryDataType {
        DirectoryDataType{
            name:   Self::CUR_DIR_NAME.to_string(),
            member: Vec::new(),
            lba,
            length: 34,
            is_hidden: false,
        }
    }

    pub fn as_flat_cur_dir(&self) -> DirectoryDataType {
        self.as_other_dir(Self::CUR_DIR_NAME, true)
    }

    pub fn as_flat_parent_dir(&self) -> DirectoryDataType {
        self.as_other_dir(Self::PARENT_DIR_NAME, true)
    }

    pub fn add_file(&mut self, file: FileDataType) -> bool {
        match self.get(file.name.as_ref()) {
            Some(_) => false,
            None    => {
                self.member.push(DataTrackType::File(file));
                return true;
            }
        }
    }

    pub fn add_dir(&mut self, dir: DirectoryDataType) -> bool {
        match self.get(dir.name.as_ref()) {
            Some(_) => false,
            None    => {
                self.member.push(DataTrackType::Directory(dir));
                return true;
            }
        }
    }

    pub fn sort(&mut self) {
        self.member.sort_by(|a, b|
            a.get_name().cmp(b.get_name())
        );

        for member in self.member.iter_mut() {
            if let DataTrackType::Directory(dir) = member {
                dir.sort();
            }
        }
    }

    pub fn calculate_lengthes(&mut self) {
        if self.is_hidden {
            self.length = 0;
        }

        else {
            let mut length = DirectoryRecord::calculate_size_for(Self::CUR_DIR_NAME, true) + DirectoryRecord::calculate_size_for(Self::PARENT_DIR_NAME, true);

            for member in self.member.iter_mut() {
                match member {
                    DataTrackType::Directory(dir) => {
                        if !dir.is_hidden {
                            length += DirectoryRecord::calculate_size_for(dir.name.as_ref(), true);
                        }
                        dir.calculate_lengthes();
                    }
                    DataTrackType::File(file)     => {
                        if !file.is_hidden {
                            length += DirectoryRecord::calculate_size_for(file.name.as_ref(), true);
                        }
                    }
                }
            }
            self.length = length;
        }
    }

    pub fn to_uppercase(&mut self) {
        if !self.is_hidden {
            self.name = self.name.to_uppercase();
            for member in self.member.iter_mut() {
                match member {
                    DataTrackType::Directory(dir) => {
                        dir.to_uppercase();
                    },
                    DataTrackType::File(file) => {
                        file.to_uppercase();
                    }
                }
            }
        }
    }

    pub fn iter(&self) -> impl Iterator<Item = &DataTrackType> {
        self.member.iter()
    }

    pub fn iter_mut(&mut self) -> impl Iterator<Item = &mut DataTrackType> {
        self.member.iter_mut()
    }

    pub fn iter_file_mut(&mut self) -> impl Iterator<Item = &mut FileDataType> {
        self.member.iter_mut().filter_map(|member| {
            match member {
                DataTrackType::Directory(_) => None,
                DataTrackType::File(file)   => Some(file),
            }
        })
    }

    pub fn iter_dir(&self) -> impl Iterator<Item = &DirectoryDataType> {
        self.member.iter().filter_map(|member| {
            match member {
                DataTrackType::Directory(dir) => Some(dir),
                DataTrackType::File(_)        => None,
            }
        })
    }

    pub fn iter_dir_mut(&mut self) -> impl Iterator<Item = &mut DirectoryDataType> {
        self.member.iter_mut().filter_map(|member| {
            match member {
                DataTrackType::Directory(dir) => Some(dir),
                DataTrackType::File(_)        => None,
            }
        })
    }

    fn get(&self, name: &str) -> Option<&DataTrackType> {
        self.member.iter().find(|entry| {
            entry.get_name() == name
        })
    }

    fn as_other_dir(&self, name: &str, is_flat: bool) -> DirectoryDataType {
        DirectoryDataType{
            name:      name.to_string(),
            member:    if is_flat {Vec::new()} else {self.member.clone()},
            lba:       self.lba,
            length:    self.length,
            is_hidden: self.is_hidden
        }
    }
}

impl LBASize for DirectoryDataType {
    fn set_lba(&mut self, lba: usize) {
        self.lba = lba;
    }

    fn get_lba(&self) -> usize {
        self.lba
    }

    fn get_length(&self) -> usize {
        round_bytes_mode2_form1(self.length)
    }

    fn is_xa_audio(&self) -> Option<bool> {
        Some(false)
    }
}

impl std::fmt::Display for DirectoryDataType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        fn print_self(f: &mut std::fmt::Formatter<'_>, indent: usize, value: &Vec<DataTrackType>) -> std::fmt::Result {
            for member in value.iter() {
                match member {
                    DataTrackType::Directory(dir) => {
                        write!(f, "{:indent$}D: {} {{\n", "", dir.name, indent=indent)?;
                        print_self(f, indent + 4, &dir.member)?;
                        write!(f, "{:indent$}}}\n", "", indent=indent)?;
                    }
                    DataTrackType::File(file)     => {
                        write!(f, "{:indent$}F: {}\n", "", file.name, indent=indent)?;
                    }
                }
            }

            Ok(())
        }

        print_self(f, 0, &self.member)
    }
}

#[derive(Clone)]
pub struct FileDataType {
    pub name:              String,
        file_path:         String,
        content:           FileContent,
        lba:               usize,
        padded_lba_length: Option<usize>,
    pub is_hidden:         bool,
}

impl FileDataType {
    pub fn new(name: &str, content: FileContent) -> FileDataType {
        FileDataType{name: name.to_string(), file_path: String::new(), content, lba: 0usize, padded_lba_length: None, is_hidden: false}
    }

    pub fn process_content(&mut self) -> Result<(), Error> {
        match &mut self.content {
            FileContent::Request(file) => {
                let result = std::mem::take(file).process()?;
            
                self.content   = FileContent::Loaded(result.0);
                self.file_path = result.1;

                if let Some(padded_lba_length) = self.padded_lba_length {
                    let lba_length = LBASize::default_get_length_lba(self)?;
                    if lba_length > padded_lba_length {
                        return Err(Error::GenericError(format!("{} padded lba length of {} exceeds calculated size of {}", self.file_path, padded_lba_length, lba_length)));
                    }
                }

                Ok(())
            },
            FileContent::Loaded(_) => Ok(())
        }
    }

    pub fn to_uppercase(&mut self) {
        if !self.is_hidden {
            self.name = self.name.to_uppercase();
        }
    }

    pub fn get_data_ref(&self) -> Result<&ProcessedFile, Error> {
        match &self.content {
            FileContent::Request(_)   => Err(Error::GenericError("Accessing raw data not possible for reuqested file".to_owned())),
            FileContent::Loaded(data) => Ok(data)
        }
    }

    pub fn get_path(&self) -> &String {
        match &self.content {
            FileContent::Request(file) => file.get_path(),
            FileContent::Loaded(_)     => &self.file_path,
        }
    }

    pub fn set_lba_padding(&mut self, padded_lba_length: usize) {
        if padded_lba_length > 0 {
            self.padded_lba_length = Some(padded_lba_length);
        }
    }
}

impl LBASize for FileDataType {
    fn set_lba(&mut self, lba: usize) {
        self.lba = lba;
    }

    fn get_lba(&self) -> usize {
        self.lba
    }

    fn get_length(&self) -> usize {
        match &self.content {
            FileContent::Request(_)   => 0,
            FileContent::Loaded(data) => {
                
                data.len()
            },
        }
    }

    fn get_length_lba(&self) -> Result<usize, Error> {
        if let Some(padded_lba_length) = self.padded_lba_length {
            Ok(padded_lba_length)
        }

        else {
            LBASize::default_get_length_lba(self)
        }
    }
    
    fn is_xa_audio(&self) -> Option<bool> {
        match &self.content {
            FileContent::Request(_)      => None,
            FileContent::Loaded(content) => Some(content.is_xa_audio()),
        }
    }
}

#[derive(Clone)]
pub enum FileContent {
    Request(RequestedFile),
    Loaded(ProcessedFile),
}

impl FileContent {
    pub fn new_raw_data(path: String) -> FileContent {
        FileContent::Request(RequestedFile::new_raw_data(path))
    }

    pub fn new_xa_audio(path: String) -> FileContent {
        FileContent::Request(RequestedFile::new_xa_audio(path))
    }
}

pub struct CUEDesc {
    pub output_file:     String,
    pub output_bin_file: String
}

impl CUEDesc {
    pub fn new() -> CUEDesc {
        CUEDesc{output_file: String::new(), output_bin_file: String::new()}
    }

    pub fn deconstruct(self) -> (String, Vec<Specifier>) {
        (self.output_file, vec!(Specifier::File{path: self.output_bin_file, format: Format::Binary}))
    }
}
