pub mod types;
pub mod psx;
pub mod cd_desc;
pub mod lba_map;

use cdtypes::{Error, types::{*, time::Time}};
use types::EncodeBehaviour;
use std::io::Write;

pub fn write_sectors(mut file: std::fs::File, sectors: Vec<types::IntermediateSector>) -> Result<(), Error> {
    let mut time = Time::cd_start();

    for sector in sectors {
        file.write(&sector.encode(time.clone(), EncodeBehaviour::Mode0AsMode2From1)?)?;
        time.add_sector()?;
    }

    Ok(())
}

pub fn write_cue(file: std::fs::File, cue_specs: Vec<cue::Specifier>) -> Result<(), Error> {
    cue::writer::write(file, cue_specs)
}

pub fn write_lba(file: &mut dyn std::io::Write, lba_map: lba_map::EntryVec, bin_path: &str, cue_path: &str) -> Result<(), Error> {
    macro_rules! write_formatted_line {
        ($file:ident, $type:expr, $name:expr, $length:expr, $lba:expr, $time_code:expr, $bytes:expr, $file_name:expr) => {
            writeln!($file, "\t{:<TYPE_WIDTH$}{:<NAME_WIDTH$}{:<LENGTH_WIDTH$}{:<LBA_WIDTH$}{:<TIMECODE_WIDTH$}{:<BYTES_WIDTH$}{}", $type, $name, $length, $lba, $time_code, $bytes, $file_name)
        };
    }
    const TYPE_WIDTH: usize     = 6;
    const NAME_WIDTH: usize     = 17;
    const LENGTH_WIDTH: usize   = 10;
    const LBA_WIDTH: usize      = 10;
    const TIMECODE_WIDTH: usize = 12;
    const BYTES_WIDTH: usize    = 10;

    fn print_dir(file: &mut dyn std::io::Write, dir: lba_map::DirectoryDataType) -> Result<(), Error> {
        let is_hidden = dir.is_hidden;

        write_formatted_line!(file, if is_hidden {"!Dir"} else {"Dir"}, dir.name, dir.lba_length, dir.lba, "<none>", dir.size, "")?;
        for entry in dir.entries {
            match entry {
                lba_map::Entry::Directory(dir)  => print_dir(file, dir)?,
                lba_map::Entry::File(file_info) => print_file(file, file_info)?,
            }
        }
        write_formatted_line!(file, "End", dir.name, "", "", "", "", "")?;
        Ok(())
    }

    fn print_file(file: &mut dyn std::io::Write, file_info: lba_map::FileDataType) -> Result<(), Error> {
        write_formatted_line!(file, if file_info.is_hidden {"!File"} else {"File"}, file_info.name, file_info.lba_length, file_info.lba, "<none>", file_info.size, file_info.file_path)?;
        Ok(())
    }

    writeln!(file, "File LBA log generated by PSXCDGEN v0.5.0\n")?;
    writeln!(file, "Image bin file: {}", bin_path)?;
    writeln!(file, "Image cue file: {}", cue_path)?;
    writeln!(file, "\nFile System:\n\n")?;

    write_formatted_line!(file, "Type", "Name", "Length", "LBA", "Time Code", "Bytes", "File")?;

    for entry in lba_map {
        match entry {
            lba_map::Entry::Directory(dir)  => print_dir(file, dir)?,
            lba_map::Entry::File(file_info) => print_file(file, file_info)?,
        }
    }

    Ok(())
}