pub mod helper;

use super::helper::{copy_data, fill_sectors, fill_sectors_upto, write_cdxa_data, write_cdxa_audio};
use super::super::{types::{ProcessedFile, builder::*, IntermediateSector}, cd_desc::{DataTrackType, DirectoryDataType, LBASize, PVD as PVDDesc}, lba_map};
use cdtypes::{
    Error, 
    types::{
        *,
        pvd::{PrimaryVolumeDescriptor, VolumeDescriptorTerminator}, 
        cdstring::{AString, CDStringValidator, DStringValidator, DString}, 
        date::{Date, SmallDate}, 
        dir_record::DirectoryRecord, 
        helper::sector_count_mode2_form1, 
        lsb_msb::*, 
        path_table::{PathTableL, PathTableB}, 
        sector::{AudioSample, Mode2Form1},
    }
};
use std::{fs::OpenOptions, io::{Read, Seek, SeekFrom}};

const PVD_IDX:usize = 16;
const VDT_IDX:usize = 16 + 1;

pub struct PathTableLayout {
    pub table_1: usize,
    pub table_2: usize,
    pub table_3: usize,
    pub table_4: usize,
    pub size:    usize
}

pub struct ContentLayout {
    pub root_lba:   usize,
    pub path_table: PathTableLayout
}

pub fn get_cd_header_size() -> usize {
    VDT_IDX + 1
}

pub fn write_licence(licence_path: String, sectors: &mut Vec<IntermediateSector>) -> Result<(), Error> {
    fn read_licence_text(file: &mut std::fs::File) -> Result<[u8; Mode2Form1::DATA_SIZE], Error> {
        let mut data = [0u8; Mode2Form1::DATA_SIZE];

        file.seek(SeekFrom::Start(0x2488))?;
        file.read_exact(&mut data)?;

        for i in 0..(data.len()-4) {
            if data[i..i+4].starts_with(&[b'S', b'o', b'n', b'y']) {
                data[i]   = b'J';
                data[i+1] = b'a';
                data[i+2] = b'b';
                data[i+3] = b'y';
            }
        }

        Ok(data)
    }

    fn set_read_licence_logo(file: &mut std::fs::File) -> Result<(), Error> {
        file.seek(SeekFrom::Start(0x2DA8))?;
        Ok(())
    }

    fn read_licence_logo(file: &mut std::fs::File) -> Result<[u8; Mode2Form1::DATA_SIZE], Error> {
        let mut data      = [0u8; Mode2Form1::DATA_SIZE];

        file.read_exact(&mut data)?;
        file.seek(SeekFrom::Current(0x120))?;

        Ok(data)
    }

    let mut licence_file = OpenOptions::new().read(true).open(licence_path)?;

    for i in 0..=15 {
        if let Some(sector) = sectors.get_mut(i) {
            *sector = {
                match i {
                     0..=3  => EmptySectorBuilder::CDXAData(),
                     4      => {
                        let sector = IntermediateSector::CDXAData(read_licence_text(&mut licence_file)?, SubHeaderBuilder::DataDefault(false));
                        set_read_licence_logo(&mut licence_file)?;

                        sector
                     },
                     5..=11 => {
                        IntermediateSector::CDXAData(read_licence_logo(&mut licence_file)?, SubHeaderBuilder::DataDefault(false))
                     },
                    12..=15 => EmptySectorBuilder::CDXAAudio(),
                    _       => {
                        return Err(Error::GenericError("SystemArea can't exceed 16 sectors".to_owned()));
                    }
                }
            };
        }

        else {
            return Err(Error::GenericError(format!("Couldn't access sector {} of System Area", i)));
        }
    }
    Ok(())
}

pub fn write_pvd(mut desc: PVDDesc, sectors: &mut Vec<IntermediateSector>, content_layout: ContentLayout) -> Result<(), Error> {
    fn write_root_dir_record(raw_pvd: &mut PrimaryVolumeDescriptor, lba: usize) {
        let root_dir = raw_pvd.get_root_dir_record_mut();
    
        write_dir_record_entry(&DataTrackType::Directory(DirectoryDataType::new_pvd_root_dir(lba)), root_dir, false);
    }

    fn write_path_table(raw_pvd: &mut PrimaryVolumeDescriptor, layout: PathTableLayout) {
        raw_pvd.path_table_size.write(layout.size as u32);
        raw_pvd.path_table_1.write(layout.table_1 as u32);
        raw_pvd.path_table_2.write(layout.table_2 as u32);
        raw_pvd.path_table_3.write(layout.table_3 as u32);
        raw_pvd.path_table_4.write(layout.table_4 as u32);
    }

    //Make sure we have enough space and make all names upper case
    fill_sectors_upto(sectors, VDT_IDX + 1);
    desc.to_uppercase();

    let     sectors_size = sectors.len();
    let     now_date     = Date::now();
    let mut raw_pvd      = PrimaryVolumeDescriptor::new();

    raw_pvd.system_id = AString::from_str("PLAYSTATION")?;
    if desc.volume_identifier.len() > 8 {
        println!("Warning: PSX might not support volume identifiers bigger then 8 chars")
    }
    raw_pvd.volume_id       = DString::from_str(desc.volume_identifier.as_str())?;
    raw_pvd.publisher_id    = AString::from_str(desc.publisher.as_str())?;
    raw_pvd.data_preparer   = AString::from_str(desc.data_preparer.as_str())?;
    raw_pvd.app_id          = AString::from_str("PLAYSTATION")?;
    raw_pvd.vol_create_time = now_date;
    raw_pvd.vol_space_size.write((sectors_size - 2) as u32);
    raw_pvd.cd_xa_id        = ['C' as u8, 'D' as u8, '-' as u8, 'X' as u8, 'A' as u8, '0' as u8, '0' as u8, '1' as u8];
    write_root_dir_record(&mut raw_pvd, content_layout.root_lba);
    write_path_table(&mut raw_pvd, content_layout.path_table);
    
    if let Some(sector) = sectors.get_mut(PVD_IDX) {
        *sector = SectorBuilder::CDXAData(raw_pvd, SubHeaderBuilder::VolumeDescriptorDefault(false));
        
        if let Some(sector) = sectors.get_mut(VDT_IDX) {

            *sector = SectorBuilder::CDXAData(VolumeDescriptorTerminator::new(), SubHeaderBuilder::VolumeDescriptorDefault(true));
            Ok(())
        }
        
        else {
            Err(Error::GenericError(String::from("Failed padding sectors to VDT")))
        }
    }

    else {
        Err(Error::GenericError(String::from("Failed padding sectors to PVD")))
    }
}

pub fn write_content(mut root_dir: DirectoryDataType, sectors: &mut Vec<IntermediateSector>, lba_map: Option<&mut lba_map::EntryVec>) -> Result<ContentLayout, Error> {
    let path_table_size = {
        let lba_size = calculate_path_table_size(&root_dir);
        if lba_size > Mode2Form1::DATA_SIZE {
            println!("Warning: Path Table has size of {} bytes but PSX might only support Path Tables of one sector data size", lba_size);
        }
        lba_size
    };
    let start_idx         = sectors.len() + (path_table_size_sector_count(path_table_size)*4);
    let new_sector_amount = arrange_content(&mut root_dir, start_idx)?;
    
    if let Some(lba_map) = lba_map {
        append_lba_for_data_track(lba_map, &root_dir)?;
    }

    root_dir.sort();
    let layout = write_path_tables(&root_dir, sectors, path_table_size)?;
    write_arranged_content(root_dir, sectors, new_sector_amount)?;

    println!("I have {} sectors because I added {} new sectors (Start: {}, PathTableSize: {})", sectors.len(), new_sector_amount, start_idx, path_table_size);
    Ok(ContentLayout{root_lba: start_idx, path_table: layout})
}

pub fn append_audio(pcm: Vec<AudioSample>, sectors: &mut Vec<IntermediateSector>) -> Result<(), Error> {
    let mut pcm = &pcm[0..];

    while pcm.len() > 0 {
        let mut pcm_sector = [AudioSample::default(); sector::Audio::SAMPLE_SIZE];

        pcm = copy_data(&mut pcm_sector, pcm);
        sectors.push(IntermediateSector::Audio(pcm_sector));
    }

    Ok(())
}

pub fn decorate_names(root_dir: &mut DirectoryDataType) -> Result<(), Error> {
    fn add_version_number(name: &mut String) -> Result<(), Error> {
        if let Some(pos) = name.rfind(';') {
            let version_str = &name[(pos + 1)..];

            if !version_str.is_empty() {
                if let Err(_) = version_str.parse::<u16>() {
                    return Err(Error::GenericError(format!("{} does not has a valid version number", name)));
                }
            }

            else {
                name.push('1');
            }
        }

        else {
            name.push_str(";1");
        }

        Ok(())
    }

    fn get_file_name_extension(string: &String) -> (&str, &str) {
        let mut name = string.as_ref();
        let mut ext  = "";

        if let Some(dot) = string.rfind('.') {
            name = &string[0..dot];
            if let Some(semi) = string.rfind(';') {
                ext = &string[(dot + 1)..semi];
            }
        }

        (name, ext)
    }

    root_dir.to_uppercase();
    for entry in root_dir.iter_mut() {
        match entry {
            DataTrackType::Directory(dir) => {
                DStringValidator::is_valid_ascii(dir.name.as_bytes())?;
                if dir.name.len() > 8 {
                    return Err(Error::GenericError(format!("Directory name {} must not have more then 8 characters", dir)));
                }
                decorate_names(dir)?;
            },
            DataTrackType::File(file)     => {
                add_version_number(&mut file.name)?;
                let (name, ext) = get_file_name_extension(&file.name);
                if ext.is_empty() {
                    return Err(Error::GenericError(format!("Missing file extension for {}", file.name)));
                }

                if name.len() > 8 {
                    return Err(Error::GenericError(format!("File name {} must not have more then 8 characters", name)));
                }

                if ext.len() > 3 {
                    return Err(Error::GenericError(format!("File extension {} for file {} must not have more then 3 characters", ext, name)));
                }

                DStringValidator::is_valid_ascii(name.as_bytes())?;
                DStringValidator::is_valid_ascii(ext.as_bytes())?;
            }
        }
    }

    Ok(())
}

fn append_lba_for_data_track(lba_map: &mut lba_map::EntryVec, root_dir: &DirectoryDataType) -> Result<(), Error> {
    fn append_dir(dir: &DirectoryDataType, map: &mut Vec<lba_map::Entry>) -> Result<(), Error> {
        for element in dir.iter() {
            match element {
                DataTrackType::Directory(dir) => {
                    let mut dir_entry = lba_map::DirectoryDataType::new(dir.name.clone(), dir, dir.is_hidden)?;
                    append_dir(dir, &mut dir_entry.entries)?;

                    map.push(lba_map::new_dir_entry(dir_entry));
                },
                DataTrackType::File(file)     => map.push(lba_map::new_file_entry(file.name.clone(), file.get_path().clone(), file, file.is_hidden)?),
            }
        }
        
        Ok(())
    }

    append_dir(root_dir, lba_map)
}

fn arrange_content(root_dir: &mut DirectoryDataType, start_lba: usize) -> Result<usize, Error> {
    fn arrange_files(dir: &mut DirectoryDataType, mut lba: usize) -> Result<usize, Error> {
        for file in dir.iter_file_mut() {
            lba = file.set_lba_ret(lba)?;
        }

        Ok(lba)
    }

    fn arrange_dirs(dir: &mut DirectoryDataType, mut lba: usize) -> Result<usize, Error> {
        for dir in dir.iter_dir_mut() {
            if !dir.is_hidden {
                lba = dir.set_lba_ret(lba)?;
            }
        }

        Ok(lba)
    }

    fn arrange(dir: &mut DirectoryDataType, mut lba: usize) -> Result<usize, Error> {
        lba = arrange_files(dir, lba)?;
        lba = arrange_dirs(dir, lba)?;

        for dir in dir.iter_dir_mut() {
            lba = arrange(dir, lba)?;
        }

        Ok(lba)
    }

    fn _alt_arrange(dir: &mut DirectoryDataType, mut lba: usize) -> Result<usize, Error> {
        for entry in dir.iter_mut() {
            match entry {
                DataTrackType::Directory(dir) => {
                    lba = dir.set_lba_ret(lba)?;
                    lba = _alt_arrange(dir, lba)?;
                },
                DataTrackType::File(file)     => lba = file.set_lba_ret(lba)?,
            }
        }

        Ok(lba)
    }

    root_dir.calculate_lengthes();

    //Root is always sorted file first
    let mut lba = start_lba;
    
    lba = root_dir.set_lba_ret(lba)?;
    lba = arrange(root_dir, lba)?;

    Ok(lba - start_lba)
}

fn write_dir_record_entry(entry: &DataTrackType, dir_rec: &mut DirectoryRecord, has_system_use: bool) {
    enum XASystemUseType
    {
        Directory,
        DataFile,
    }

    unsafe {dir_rec.new(entry.get_name().as_ref(), has_system_use)};

    dir_rec.data_block_number.write(entry.get_lba() as u32);
    dir_rec.data_size.write(entry.get_length() as u32);
    dir_rec.time_stamp = SmallDate::now();

    let entry_type = {
        match entry {
            DataTrackType::Directory(_) => {
                dir_rec.set_directory();
                XASystemUseType::Directory
            }

            DataTrackType::File(_)      => {
                dir_rec.set_file();
                XASystemUseType::DataFile
            }
        }
    };

    if let Some(cdxa_entry) = dir_rec.get_cdxa_system_use_mut() {
        match entry_type {
            XASystemUseType::Directory => {
                cdxa_entry.file_attribute.set_mode2();
                cdxa_entry.file_attribute.set_directory();
            }
            XASystemUseType::DataFile  => {
                cdxa_entry.file_attribute.set_mode2();
            }
        }
    }
}

fn path_table_size_sector_count(bytes: usize) -> usize {
    sector_count_mode2_form1(bytes)
}

fn write_path_tables(arranged_root_dir: &DirectoryDataType, sectors: &mut Vec<IntermediateSector>, size: usize) -> Result<PathTableLayout, Error> {
    type PseudoArray = (Vec<u8>, usize);

    fn push_table_entry(raw_buffer: &mut PseudoArray, name: &str, lba: usize, parent_table_id: usize) -> Result<(), Error> {
        let entry_size = PathTableL::calculate_size_for(name);
        if raw_buffer.1 + entry_size > raw_buffer.0.len() {
            return Err(Error::GenericError(format!("Out of memory for Table Path Entry!")));
        }

        let path_entry = unsafe{
            let entry = std::mem::transmute::<&mut u8, &mut PathTableL>(&mut raw_buffer.0[raw_buffer.1]);

            entry.new(name);
            entry
        };

        path_entry.directory_logical_block.write(lba as u32);
        path_entry.parent_table_id.write(parent_table_id as u16);
        raw_buffer.1 += entry_size;

        Ok(())
    }

    fn push_all_entries(raw_buffer: &mut PseudoArray, dir: &DirectoryDataType, mut parent: usize) -> Result<(), Error> {
        if !dir.is_hidden {
            for dir in dir.iter_dir() {
                if !dir.is_hidden {
                    push_table_entry(raw_buffer, dir.name.as_ref(), dir.get_lba(), parent)?;
                }
            }

            for dir in dir.iter_dir() {
                if !dir.is_hidden {
                    parent += 1;
                    push_all_entries(raw_buffer, dir, parent)?;
                }
            }
        }
        Ok(())
    }

    fn write_table(sectors: &mut Vec<IntermediateSector>, mut idx: usize, length_lba: usize, data: &[u8]) -> Result<usize, Error> {
        for _ in 0..2 {
            write_cdxa_data(sectors, idx, &data, SubHeaderBuilder::DataDefault(false))?;
            idx += length_lba;
        }

        Ok(idx)
    }

    fn transform_table(data_l: &mut [u8]) {
        let mut idx = 0;
        while idx < data_l.len() {
            let org_value = unsafe {std::mem::transmute::<&u8,     &PathTableL>(&data_l[idx])};
            let new_value = unsafe {std::mem::transmute::<&mut u8, &mut PathTableB>(&mut data_l[idx])};

            new_value.directory_logical_block.write(org_value.directory_logical_block.read());
            new_value.parent_table_id.write(org_value.parent_table_id.read());

            idx += new_value.get_size();
        }
    }

    let mut raw_buffer_table = (vec![0u8; size], 0);
    
    push_table_entry(&mut raw_buffer_table, "\x00", arranged_root_dir.get_lba(), 1)?;
    push_all_entries(&mut raw_buffer_table, arranged_root_dir, 1)?;

    let mut table_idx = sectors.len();
    let size_lba      = path_table_size_sector_count(size);

    let layout = PathTableLayout{table_1: table_idx, table_2: table_idx + size_lba, table_3: table_idx + (2*size_lba), table_4: table_idx + (3*size_lba), size: size};

    fill_sectors(sectors, size_lba*4);
    table_idx = write_table(sectors, table_idx, size_lba, &raw_buffer_table.0)?;
    transform_table(&mut raw_buffer_table.0);
    write_table(sectors, table_idx, size_lba, &raw_buffer_table.0)?;

    Ok(layout)
}

fn write_arranged_content(arranged_root_dir: DirectoryDataType, sectors: &mut Vec<IntermediateSector>, amount: usize) -> Result<(), Error> {
    fn write_dir(dir: &DirectoryDataType, parent: &DirectoryDataType, sectors: &mut Vec<IntermediateSector>) -> Result<(), Error> {
        fn set_sector(data: [u8; Mode2Form1::DATA_SIZE], is_last: bool, lba: usize, sectors: &mut Vec<IntermediateSector>) -> Result<(), Error> {
            if let Some(sector) = sectors.get_mut(lba) {
                *sector = IntermediateSector::CDXAData(data, SubHeaderBuilder::DirectoryRecordDefault(is_last));
                Ok(())
            }

            else {
                Err(Error::NoSectorForLBA(lba))
            }
        }

        if dir.is_hidden {
            Ok(())
        }

        else {
            let entry_length      = dir.get_length_lba()?;
            let mut sector_buffer = [0u8; Mode2Form1::DATA_SIZE];
            let mut sector_idx    = 0usize;
            let mut lba           = dir.get_lba();
            let mut write_entry   = |entry: &DataTrackType| -> Result<(), Error> {
                if entry.is_hidden() {
                    Ok(())
                }

                else {
                    let name = entry.get_name().as_ref();
                    let size = DirectoryRecord::calculate_size_for(name, true);
                    if sector_idx + size > Mode2Form1::DATA_SIZE {
                        set_sector(sector_buffer, false, lba, sectors)?;

                        sector_buffer = [0u8; Mode2Form1::DATA_SIZE];
                        lba += 1;
                    }

                    let dir_rec = unsafe {std::mem::transmute::<&mut u8, &mut DirectoryRecord>(&mut sector_buffer[sector_idx])};
                    write_dir_record_entry(entry, dir_rec, true);

                    sector_idx += size;
                    Ok(())
                }
            };

            if entry_length > 1 {
                println!("Warning: Directory {} spans {} sectors and might not be supported by PSX", dir.name, entry_length);
            }

            //Write current and parent dir
            write_entry(&DataTrackType::Directory(dir.as_flat_cur_dir()))?;
            write_entry(&DataTrackType::Directory(parent.as_flat_parent_dir()))?;

            for entry in dir.iter() {
                write_entry(entry)?;
            }
            set_sector(sector_buffer, true, lba, sectors)?;
            Ok(())
        }
    }

    fn write(parent_dir: &DirectoryDataType, sectors: &mut Vec<IntermediateSector>) -> Result<(), Error> {
        for entry in parent_dir.iter() {
            match entry {
                DataTrackType::Directory(dir) => {
                    write_dir(dir, parent_dir, sectors)?;
                    write(dir, sectors)?;
                },

                DataTrackType::File(file)     => {
                    match file.get_data_ref()? {
                        ProcessedFile::Raw(raw_data)     => write_cdxa_data(sectors, file.get_lba(), raw_data.as_ref(), SubHeaderBuilder::DataDefault(false))?,
                        ProcessedFile::XAAudio(raw_data) => write_cdxa_audio(sectors, file.get_lba(), raw_data)?,
                    }
                }
            }
        }
        Ok(())
    }

    fill_sectors(sectors, amount);

    write_dir(&arranged_root_dir, &arranged_root_dir, sectors)?;
    write(&arranged_root_dir, sectors)
}

fn calculate_path_table_size(root_dir: &DirectoryDataType) -> usize {
    if root_dir.is_hidden {
        0
    }

    else {
        let mut size = PathTableL::calculate_size_for("\x00");

        for sub_dir in root_dir.iter_dir() {
            if !sub_dir.is_hidden {
                size += PathTableL::calculate_size_for(sub_dir.name.as_ref());
                size += calculate_path_table_size(sub_dir)
            }
        }

        size
    }
}