use super::{types::IntermediateSector, cd_desc::{CDDesc, DataTrackType, DirectoryDataType, Track}, lba_map};
use cdtypes::{types::{*, time::Time}, Error};
use encoder::helper;

pub mod encoder;

pub fn encode(cd_desc: CDDesc, cue_specs: &mut Vec<cue::Specifier>) -> Result<(Vec<IntermediateSector>, Option<lba_map::EntryVec>), Error> {
    let     header_size    = encoder::get_cd_header_size();
    let mut sectors        = Vec::new();
    let mut content_layout = None;
    let mut cur_track      = 1;
    let mut cur_time       = Time::cue_start();
    let mut lba_map        = Vec::new();

    // Creates enough space for the PVD
    helper::fill_sectors_upto(&mut sectors, header_size);

    // Write Licence
    if !cd_desc.licence_path.is_empty() {
        encoder::write_licence(cd_desc.licence_path, &mut sectors)?;
    }

    for track in cd_desc.tracks {
        match track {
            Track::Data(mut dir) => {
                process_dir(&mut dir)?;
                // Decorate and verify the names for content creation
                encoder::decorate_names(&mut dir)?;
                let new_content_layout = encoder::write_content(dir, &mut sectors, Some(&mut lba_map))?;

                if content_layout.is_none() {
                    content_layout = Some(new_content_layout);
                }

                cue_specs.push(cue::Specifier::Track{number: cur_track, data_type: cue::DataType::Mode2_2352});
                cue_specs.push(cue::Specifier::Index{number: 1, time: cur_time});
            }

            Track::Audio(pcm) => {
                encoder::append_audio(pcm, &mut sectors)?;

                cue_specs.push(cue::Specifier::Track{number: cur_track, data_type: cue::DataType::Audio});
                cue_specs.push(cue::Specifier::PreGap{time: Time::cd_pregap()});
                cue_specs.push(cue::Specifier::Index{number: 1, time: cur_time});
            }
        }

        cur_track += 1;
        cur_time  = Time::from_lba(sectors.len());
    }

    if let Some(content_layout) = content_layout {
        // Now write the PVD cause write_content eventaully tells us where to start
        encoder::write_pvd(cd_desc.pvd, &mut sectors, content_layout)?;
    }
    
    else {
        return Err(Error::GenericError("No Data track found!".to_owned()));
    }

    Ok((sectors, Some(lba_map)))
}

fn process_dir(dir: &mut DirectoryDataType) -> Result<(), Error> {
    for entry in dir.iter_mut() {
        match entry {
            DataTrackType::Directory(dir) => process_dir(dir)?,
            DataTrackType::File(file)     => file.process_content()?,
        }       
    }

    Ok(())
}