pub mod builder;

use cdtypes::{Error, types::{error_correction, sector, time::Time}};
use builder::SubHeaderBuilder;
use super::super::file_helper::read_file_to_u8;

type ByteSector = [u8; sector::SECTOR_SIZE];

pub enum EncodeBehaviour {
    Normal,
    Mode0AsMode2From1
}

pub enum IntermediateSector {
    Empty,
    Audio([sector::AudioSample; sector::Audio::SAMPLE_SIZE]),
    CDData([u8; sector::Mode1::DATA_SIZE]),
    CDXAData([u8;sector::Mode2Form1::DATA_SIZE], sector::SubHeader),
    CDXAAudio([u8;sector::Mode2Form2::DATA_SIZE], sector::SubHeader),
}

impl IntermediateSector {
    pub fn encode(self, time: Time, behaviour: EncodeBehaviour) -> Result<ByteSector, Error> {
        match self {
            Self::Empty                      =>
                match behaviour {
                    EncodeBehaviour::Normal            => Ok(Self::encode_mode0(time)),
                    EncodeBehaviour::Mode0AsMode2From1 => Ok(Self::encode_mode0_as_mode2_form1(time))
            },
            Self::Audio(data)                 => Ok(unsafe {std::mem::transmute::<[sector::AudioSample; sector::Audio::SAMPLE_SIZE], ByteSector>(data)}),
            Self::CDData(data)                => Ok(Self::encode_mode1(time, data)),
            Self::CDXAData(data,  sub_header) => Ok(Self::encode_mode2_form1(time, data, sub_header)),
            Self::CDXAAudio(data, sub_header) => Ok(Self::encode_mode2_form2(time, data, sub_header)),
        }
    }

    fn encode_mode0(time: Time) -> ByteSector {
        let mut sector = sector::Mode0::new();

        sector.header.set_time(time);
        unsafe {
            std::mem::transmute::<sector::Mode0, ByteSector>(sector)
        }
    }

    fn encode_mode0_as_mode2_form1(time: Time) -> ByteSector {
        Self::encode_mode2_form1(time, [0u8; sector::Mode1::DATA_SIZE], SubHeaderBuilder::DataDefault(true))
    }

    fn encode_mode1(time: Time, data: [u8; sector::Mode1::DATA_SIZE]) -> ByteSector {
        let mut sector = sector::Mode1::new();

        sector.header.set_time(time);
        sector.data = data;
        sector.finalize();

        unsafe {
            std::mem::transmute::<sector::Mode1, ByteSector>(sector)
        }
    }

    fn encode_mode2_form1(time: Time, data: [u8; sector::Mode2Form1::DATA_SIZE], sub_header: sector::SubHeader) -> ByteSector {
        let mut sector = sector::Mode2Form1::new();

        sector.header.set_time(time);
        sector.data       = data;
        sector.sub_header = sub_header;

        sector.finalize();
        unsafe {
            std::mem::transmute::<sector::Mode2Form1, ByteSector>(sector)
        }
    }

    fn encode_mode2_form2(time: Time, data: [u8; sector::Mode2Form2::DATA_SIZE], sub_header: sector::SubHeader) -> ByteSector {
        let mut sector = sector::Mode2Form2::new();

        sector.header.set_time(time);
        sector.data       = data;
        sector.sub_header = sub_header;

        sector.finalize();
        unsafe {
            std::mem::transmute::<sector::Mode2Form2, ByteSector>(sector)
        }
    }
}

#[derive(Clone)]
pub struct RequestedFile {
    path:      String,
    data_type: DataType,
}

impl RequestedFile {
    pub fn new_raw_data(file_path: String) -> RequestedFile {
        Self::new_simple(file_path, DataType::RawData)
    }

    pub fn new_xa_audio(file_path: String) -> RequestedFile {
        Self::new_simple(file_path, DataType::XAAudio)
    }

    pub fn process(self) -> Result<(ProcessedFile, String), Error> {
        let data = read_file_to_u8(self.path.as_ref())?;

        match self.data_type {
            DataType::None    => Ok((ProcessedFile::Raw(Vec::new()), self.path)),
            DataType::RawData => Ok((ProcessedFile::Raw(data),       self.path)),
            DataType::XAAudio => Ok((Self::parse_xa_audio(data),     self.path)),
        }        
    }

    pub fn get_path(&self) -> &String {
        &self.path
    }

    fn new_simple(file_path: String, data_type: DataType) -> RequestedFile {
        RequestedFile{path: file_path, data_type}
    }

    fn parse_xa_audio(data: Vec<u8>) -> ProcessedFile {
        const XAAUDIOFILE_SIZE:usize = std::mem::size_of::<XAAudioFile>();
        
        let mut data    = &data[..];
        let mut xa_data = Vec::new();

        while data.len() >= XAAUDIOFILE_SIZE {
            let content = unsafe {&*(data.as_ptr() as *const XAAudioFile)};

            xa_data.push((content.sub_header.clone(), content.data));
            data = &data[XAAUDIOFILE_SIZE..];
        }
        ProcessedFile::XAAudio(xa_data)
    }
}

impl std::default::Default for RequestedFile {
    fn default() -> RequestedFile {
        RequestedFile{path: String::new(), data_type: DataType::None}
    }
}

#[derive(Clone)]
pub enum ProcessedFile {
    Raw(Vec<u8>),
    XAAudio(Vec<(sector::SubHeader, [u8; sector::Mode2Form2::DATA_SIZE])>),
}

impl ProcessedFile {
    pub fn len(&self) -> usize {
        match self {
            Self::Raw(data)     => data.len(),
            Self::XAAudio(data) => data.len()*sector::Mode2Form2::DATA_SIZE,
        }
    }

    pub fn is_xa_audio(&self) -> bool {
        match self {
            Self::XAAudio(_) => true,
            _                => false
        }
    }
}

#[derive(Clone)]
enum DataType {
    None,
    RawData,
    XAAudio,
}

#[repr(packed(1))]
struct XAAudioFile {
    pub sub_header:       sector::SubHeader,
    pub _sub_header_copy: sector::SubHeader,
    pub data:             [u8; sector::Mode2Form2::DATA_SIZE],
    pub _edc:             error_correction::EDC
}