use psxcdgen::{*, file_helper::{read_file_to_string, open_output_file, path_to_string}, iso_writer::{psx, cd_desc::*}};
use cdtypes::Error;
use clap::{Parser};
use std::{env, path::PathBuf};

#[derive(Parser)]
#[clap(about = "Generates a PSX ISO", long_about = None)]
struct CommandLine {
    //Overrides output (not supported)
    #[clap(short='y')]
    override_output: bool,

    //Output file for the lba values (not supported)
    #[clap(long="lba")]
    lba_file: Option<PathBuf>,

    //Output header file for the lba values (not supported)
    #[clap(long="lbahead")]
    lbahead_file: Option<PathBuf>,

    //Disables iso generation (not supported)
    #[clap(long="noisogen")]
    no_isogen: bool,

    //The input XML
    input_path: PathBuf
}

struct TaskError {
    name:  String,
    error: Error,
}

impl std::fmt::Display for TaskError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{} failed with: \"{}\"", self.name, self.error)
    }
}

fn start_task<T>(name: &str, result: Result<T, Error>) -> Result<T, TaskError> {
    match result {
        Ok(value)  => Ok(value),
        Err(error) => Err(TaskError{name: name.to_owned(), error})
    }
}

fn write_desc(mut cd_desc: CDDesc, cue_desc: CUEDesc) -> Result<(), TaskError> {
    let cd_output_file            = std::mem::take(&mut cd_desc.output_file);
    let (cue_path, mut cue_specs) = cue_desc.deconstruct();
    let file                      = start_task("Open iso output file", open_output_file(cd_output_file.as_ref()))?;
    let cue_file                  = start_task("Open cue output file", open_output_file(cue_path.as_ref()))?;
    let (encoded_sector, lba_map) = start_task("Encode CD sectors",    psx::encode(cd_desc, &mut cue_specs))?;

    start_task("Writing ISO image", iso_writer::write_sectors(file, encoded_sector))?;   
    start_task("Writing CUE file",  iso_writer::write_cue(cue_file, cue_specs))?;
    if let Some(lba_map) = lba_map {
        start_task("Write LBA file", iso_writer::write_lba(&mut std::io::stdout(), lba_map, cd_output_file.as_ref(), cue_path.as_ref()))?;
    }

    Ok(())
}

fn run_main(arguments: CommandLine) -> Result<(), TaskError> {
    if !arguments.no_isogen {
        let input_path          = start_task("Opening XML file",               path_to_string(arguments.input_path))?;
        let input_content       = start_task("Reading XML file",               read_file_to_string(input_path.as_ref()))?;
        let (cd_desc, cue_desc) = start_task("Parsing XML file like mkpsxiso", xml_reader::parse_mkpsxiso(input_content))?;    

        write_desc(cd_desc, cue_desc)
    }

    else {
        Ok(())
    }
}

fn main() {
    //Only needed if running in mkpsxiso mode (Currently) the default
    fn edit_env_args() -> Vec<String> {
        let mut args = Vec::new();

        for argument in env::args() {
            let mut argument = argument.to_owned();

            if argument.len() > 2 && argument.starts_with('-') {
                argument.insert(0, '-');
            }
            args.push(argument);
        }

        args
    }

    //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    match CommandLine::try_parse_from(edit_env_args().iter()) {
        Ok(arguments) => {
            match run_main(arguments) {
                Ok(())     => println!("Succees!"),
                Err(error) => println!("{}", error)
            }
        },
        Err(error)    => println!("{}", error)
    }
}