use cdtypes::{types::sector::AudioSample, Error};
use super::{
    XMLReader,
    super::iso_writer::cd_desc::{CDDesc, CUEDesc, Track, DirectoryDataType, FileContent, FileDataType}
};
use fast_xml::events::{attributes::Attributes, Event};

macro_rules! make_tag_name {
    ($name:expr) => {
        paste::item! {
            const [< $name:upper _TAG>]:  &'static str  = $name;
            const [< $name:upper _TAG_B>]:&'static [u8] = [< $name:upper _TAG>].as_bytes();
        }
    };
}

make_tag_name!("iso_project");
make_tag_name!("track");
make_tag_name!("identifiers");
make_tag_name!("directory_tree");
make_tag_name!("dir");

struct Desc {
    cd:  CDDesc,
    cue: CUEDesc,
}

impl Desc {
    fn new() -> Desc {
        Desc{cd: CDDesc::new(), cue: CUEDesc::new()}
    }
}

pub fn parse(xml: String) -> Result<(CDDesc, CUEDesc), Error> {
    let mut desc   = Desc::new();
    let mut parser = XMLReader::new(xml.as_ref());

    loop {
        match parser.next()? {
            Event::Start(tag) => {
                match tag.name() {
                    ISO_PROJECT_TAG_B => {
                        handle_iso_project_attributes(&mut desc, tag.attributes())?;
                        handle_iso_project(&mut desc, &mut parser)?;
                    },
                    _ => (),
                }
            },
            Event::Eof => {
                return Ok((desc.cd, desc.cue));
            },
            _          => (),
        }
    }
}

fn handle_iso_project_attributes(desc: &mut Desc, attributes: Attributes) -> Result<(), Error> {
    XMLReader::for_each_attribute(attributes, |attribute| {
        match attribute.key {
            b"image_name" => {
                desc.cd.output_file      = XMLReader::get_attribute_value_str(&attribute)?.to_string();
                desc.cue.output_bin_file = desc.cd.output_file.clone();
            },
            b"cue_sheet"  => {
                desc.cue.output_file = XMLReader::get_attribute_value_str(&attribute)?.to_string();
            },
            _   => XMLReader::print_attribute_not_supported(attribute.key, ISO_PROJECT_TAG),
        }
        
        Ok(())
    })
}

fn handle_iso_project(desc: &mut Desc, parser: &mut XMLReader) -> Result<(), Error> {
    enum TrackType {
        None,
        Data,
        Audio,
    }

    loop {
        match parser.next()? {
            Event::Start(tag) => {
                match tag.name() {

                    TRACK_TAG_B => {
                        let mut track_type = TrackType::None;
                        let mut source     = String::default();

                        XMLReader::for_each_attribute(tag.attributes(), |attribute| {
                            match attribute.key {
                                b"type" => {
                                    let attribute_value = XMLReader::get_attribute_value_str(&attribute)?;
                                    match attribute_value {
                                        "data"  => track_type = TrackType::Data,
                                        "audio" => track_type = TrackType::Audio,
                                        _       => println!("Unkown track type {}", attribute_value), 
                                    }
                                },
                                b"source" => {
                                    source = XMLReader::get_attribute_value_str(&attribute)?.to_owned();
                                },
                                _ => XMLReader::print_attribute_not_supported(attribute.key, TRACK_TAG)
                            }

                            Ok(())
                        })?;

                        match track_type {
                            TrackType::None |
                            TrackType::Data  => handle_data_track(desc, parser)?,
                            TrackType::Audio => handle_audio_track(desc, source)?,
                        }
                    }
                    _        => (),
                }
            },
            Event::End(tag) => {
                match tag.name() {
                    ISO_PROJECT_TAG_B => {
                        return Ok(());
                    }
                    _ => (),
                }
            }

            Event::Eof => break,
            _          => (),
        }
    }

    Err(Error::EOF)
}

fn handle_data_track(desc: &mut Desc, parser: &mut XMLReader) -> Result<(), Error> {
    loop {
        match parser.next()? {
            Event::Start(tag) => {
                match tag.name() {
                    IDENTIFIERS_TAG_B => handle_identifiers_attributes(desc, tag.attributes())?,
                    b"license"        => {
                        if let Some(attribute) = XMLReader::find_attribute(tag.attributes(), b"file")? {
                            desc.cd.licence_path = XMLReader::get_attribute_value_str(&attribute)?.to_owned();
                        }
    
                        else {
                            println!("No file specified for licence!");
                        }
                    },
                    DIRECTORY_TREE_TAG_B => {
                        let mut dir = DirectoryDataType::new("track");
                        handle_directory(&mut dir, parser)?;

                        desc.cd.tracks.push(Track::Data(dir));
                    },
                    _                 => (),
                }
            },
            Event::End(tag) => {
                match tag.name() {
                    TRACK_TAG_B => {
                        // Push track here?
                        return Ok(());
                    },
                    _ => ()
                }
            }
            Event::Eof => break,
            _          => ()
        }
    }
    Err(Error::EOF)
}

fn handle_identifiers_attributes(desc: &mut Desc, attributes: Attributes) -> Result<(), Error> {
    XMLReader::for_each_attribute(attributes, |attribute| {
        match attribute.key {
            b"system" | b"application" => {
                let value = XMLReader::get_attribute_value_str(&attribute)?;
                if value != "PLAYSTATION" {
                    println!("System/Application Identifier \"{}\" will be ignored", value);
                }
            },
            b"volume" => {
                desc.cd.pvd.volume_identifier = XMLReader::get_attribute_value_str(&attribute)?.to_owned();
            },
            b"volume_set" | b"copyright" => {
                println!("{} with \"{}\" ignored", XMLReader::get_attribute_key_str(&attribute)?, XMLReader::get_attribute_value_str(&attribute)?);
            },
            b"publisher"                 => {
                desc.cd.pvd.publisher = XMLReader::get_attribute_value_str(&attribute)?.to_string();
            },
            b"data_preparer"             => {
                desc.cd.pvd.data_preparer = XMLReader::get_attribute_value_str(&attribute)?.to_string();
            },
            _                            => {
                XMLReader::print_attribute_not_supported(attribute.key, IDENTIFIERS_TAG);
            }
        }

        Ok(())
    })
}

fn handle_directory(dir: &mut DirectoryDataType, parser: &mut XMLReader) -> Result<(), Error> {
    enum FileType {
        None,
        Data,
        Xa,
    }

    loop {
        match parser.next()? {
            Event::Start(tag) => {
                match tag.name() {
                    b"file"   => {
                        let mut file_type   = FileType::None;
                        let mut file_name   = String::default();
                        let mut file_source = String::default();

                        XMLReader::for_each_attribute(tag.attributes(), |attribute| {
                            match attribute.key {
                                b"name"   => {
                                    file_name   = XMLReader::get_attribute_value_str(&attribute)?.to_string();
                                    Ok(())
                                },
                                b"source" => {
                                    file_source = XMLReader::get_attribute_value_str(&attribute)?.to_string();
                                    Ok(())
                                },
                                b"type" => {
                                    if let Ok(r#type) = XMLReader::get_attribute_value_str(&attribute) {
                                        if r#type == "data" {
                                            file_type = FileType::Data;
                                            Ok(())
                                        }

                                        else if r#type == "xa" {
                                            file_type = FileType::Xa;
                                            Ok(())
                                        }

                                        else {
                                            println!("File type: \"{}\" not supported yet", r#type);
                                            Err(Error::NotImplemented)
                                        }
                                    }

                                    else {
                                        Err(Error::GenericError("Attribute \"type\" needs to be either data or xa".to_owned()))
                                    }
                                }
                                _ => Ok(()),
                            }
                        })?;

                        match file_type {
                            FileType::None | FileType::Data => {
                                let mut file = FileDataType::new(file_name.as_ref(), FileContent::new_raw_data(file_source));

                                if file.name == "SCES_003.90" {
                                    file.set_lba_padding(0);
                                    file.is_hidden = false;
                                }
                                dir.add_file(file);
                            },
                            FileType::Xa => {
                                dir.add_file(FileDataType::new(file_name.as_ref(), FileContent::new_xa_audio(file_source)));
                            }
                        }
                    },
                    DIR_TAG_B => {
                        if let Some(dir_name) = XMLReader::find_attribute(tag.attributes(), b"name")? {
                            let mut new_dir = DirectoryDataType::new(XMLReader::get_attribute_value_str(&dir_name)?);

                            if new_dir.name == "XB" {
                                new_dir.is_hidden = false;
                            }
                            handle_directory(&mut new_dir, parser)?;
                            dir.add_dir(new_dir);
                        }

                        else {
                            return Err(Error::GenericError("Dir requires name".to_owned()));
                        }
                    },
                    _ => (),
                }
            },
            Event::End(tag) => {
                match tag.name() {
                    DIRECTORY_TREE_TAG_B | DIR_TAG_B => {
                        return Ok(());
                    },
                    _ => ()
                }
            }
            Event::Eof => break,
            _          => ()
        }
    }
    Err(Error::EOF)
}

fn handle_audio_track(desc: &mut Desc, file: String) -> Result<(), Error> {
    fn convert_into_16(data: wav::bit_depth::BitDepth) -> Result<Vec<AudioSample>, Error> {
        match data.try_into_sixteen() {
            Ok(data) => {
                let mut samples = Vec::new();

                for (left, right) in data.iter().step_by(2).zip(data.iter().skip(1).step_by(2)) {
                    samples.push(AudioSample::new(*left, *right))
                }

                Ok(samples)
            },
            Err(_)   => Err(Error::GenericError("Couldn't convert samples to 16bit".to_owned())),
        }
    }

    let (header, data) = wav::read(&mut std::fs::File::open(file)?)?;

    if header.audio_format != wav::header::WAV_FORMAT_PCM {
        return Err(Error::GenericError("Only WAV PCM is supported for Audio tracks".to_owned()));
    }

    if header.sampling_rate != 44_100 {
        return Err(Error::GenericError("Only a sampling rate of 44.1kHz is supported for Audio tracks".to_owned()));
    }

    desc.cd.tracks.push(Track::Audio(convert_into_16(data)?));
    Ok(())
}