use super::{layout::Layout, File, FileSystemMap};
use super::super::encoder::LengthCalculatorFunction;
use std::path::PathBuf;
use no_comment::{IntoWithoutComments as _, languages};
use tool_helper::{Error, format_if_error, read_file};

pub type LBANameVec = Vec<String>;

const COMPRESSION_LEVEL:u32 = 16;

#[repr(packed)]
struct OverlayHeader {
    _start_adr: u32,
    lba_count: u16,
}

impl OverlayHeader {
    pub fn read_lba_count(&self) -> usize {
        u16::from_le(self.lba_count) as usize
    }
}

#[repr(packed)]
struct LBAEntry {
    lba:     u16,
    sectors: u16,
}

impl LBAEntry {
    pub fn write_entry(&mut self, lba: u16, sectors: u16) {
        let lba     = lba.to_le_bytes();
        let sectors = sectors.to_le_bytes();

        self.lba     = u16::from_ne_bytes(lba);
        self.sectors = u16::from_ne_bytes(sectors);
    }
}

pub fn load_from(file_name: &str, file_path: PathBuf, lba_source: PathBuf) -> Result<File, Error> {
    let content      = load_content(&file_path)?;
    let lba_names    = load_lba_names(lba_source)?;
    let content_size = format_if_error!(tool_helper::compress::lz4(&content, COMPRESSION_LEVEL), "Compressing {} failed with \"{error_text}\"", file_path.to_string_lossy())?.len();
    
    Ok(File::new_overlay(file_name, content, lba_names, content_size)?)
}

pub fn update_content(content: &mut Vec<u8>, lba_names: &LBANameVec, file_map: &FileSystemMap, length_func: LengthCalculatorFunction) -> Result<Vec<u8>, Error> {
    let (lba_header, lba_count) = skip_to_lba_header(content);
    let  lba_header             = unsafe{std::slice::from_raw_parts_mut(lba_header.as_mut_ptr() as *mut LBAEntry, lba_count)};
    
    let mut idx = 0;
    for lba_name in lba_names {
        if let Some(file) = file_map.get(lba_name) {
            let (lba, sector_count) = (file.borrow().get_track_rel_lba(), length_func(&Layout::File(file.clone())).sectors);

            if idx >= lba_count {
                return Err(Error::from_text(format!("Trying to write more LBAs then there is space!")));
            }

            lba_header[idx].write_entry(lba as u16, sector_count as u16);
            idx += 1;
        }

        else {
            return Err(Error::from_text(format!("Could not locate file {} on disk", lba_name)));
        }
    }

    Ok(tool_helper::compress::lz4(content, COMPRESSION_LEVEL)?)
}

fn load_content(file_path: &PathBuf) -> Result<Vec<u8>, Error> {
    let mut content = read_file(&file_path)?;

    if content.len() < std::mem::size_of::<OverlayHeader>() {
        return Err(Error::from_text(format!("Overlay {} has no header!", file_path.to_string_lossy())));
    }

    let (lba_header, lba_count) = skip_to_lba_header(&mut content);
    if lba_header.is_empty() {
        return Err(Error::from_text(format!("LBA header of overlay {} is smaller then {} elements", file_path.to_string_lossy(), lba_count)));
    }

    let mut count = 0;
    for byte in lba_header {
        *byte = count as u8;
        count += 1;
    }

    Ok(content)
}

fn skip_to_lba_header(content: &mut Vec<u8>) -> (&mut [u8], usize) {
    let overlay_header_size = std::mem::size_of::<OverlayHeader>();

    let lba_count       = unsafe{std::mem::transmute::<&u8, &OverlayHeader>(&content[0])}.read_lba_count();
    let lba_header_size = lba_count*std::mem::size_of::<LBAEntry>();

    (&mut content[overlay_header_size..(overlay_header_size+lba_header_size)], lba_count)
}

fn load_lba_names(lba_source: PathBuf) -> Result<LBANameVec, Error> {
    const LBA_DECLARATION:&'static str = "__jabyengine_request_lba_for";
    fn get_part_of_interest(file: String, lba_source: &PathBuf) -> Result<String, Error> {
        const LBA_START_POINT:&'static str = "__jabyengine_start_lba_request";
        const LBA_END_POINT:&'static str   = "__jabyengine_end_lba_request";

        let start = file.find(LBA_START_POINT).ok_or_else(|| {
            Error::from_text(format!("No LBA start found in file {}", lba_source.to_string_lossy()))
        })? + LBA_START_POINT.len();
    
        let end = file.find(LBA_END_POINT).ok_or_else(|| {
            Error::from_text(format!("No LBA end found in file {}", lba_source.to_string_lossy()))
        })?;

        Ok(file[start..end].to_owned())
    }

    let     file      = std::fs::read_to_string(&lba_source)?.chars().without_comments(languages::c()).collect::<String>();
    let     file      = get_part_of_interest(file, &lba_source)?;
    let mut lba_names = Vec::new();

    for pair in file.split(LBA_DECLARATION) {
        if let Some(mut start) = pair.find('"') {
            start += 1;
            if let Some(end) = pair[start..].find('"') {
                let mut lba_name = pair[start..(start + end)].to_owned();
                if lba_name.find(';').is_none() {
                    lba_name.push_str(";1");
                }
                lba_names.push(lba_name);
            }
        }
    }

    Ok(lba_names)
}